from collections import ChainMap
from copy import copy
from functools import partial

from PySide6.QtCore import (
    Qt,
    Signal,
    Slot,
)
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QFrame,
    QGroupBox,
    QLabel,
    QScrollArea,
    QSpinBox,
    QStyledItemDelegate,
    QTableWidget,
    QTableWidgetItem,
)

from plover import _
from plover.config import MINIMUM_UNDO_LEVELS, MINIMUM_TIME_BETWEEN_KEY_PRESSES
from plover.misc import expand_path, shorten_path
from plover.registry import registry

from plover.gui_qt.config_window_ui import Ui_ConfigWindow
from plover.gui_qt.config_file_widget_ui import Ui_FileWidget
from plover.gui_qt.utils import WindowStateMixin


class NopeOption(QLabel):
    valueChanged = Signal(bool)

    def __init__(self):
        super().__init__()
        # i18n: Widget: “NopeOption” (empty config option message,
        # e.g. the machine option when selecting the Treal machine).
        self.setText(_("Nothing to see here!"))

    def setValue(self, value):
        pass


class BooleanOption(QCheckBox):
    valueChanged = Signal(bool)

    def __init__(self):
        super().__init__()
        self.stateChanged.connect(lambda: self.valueChanged.emit(self.isChecked()))

    def setValue(self, value):
        self.setChecked(value)


class IntOption(QSpinBox):
    def __init__(self, maximum=None, minimum=None):
        super().__init__()
        if maximum is not None:
            self.setMaximum(maximum)
        if minimum is not None:
            self.setMinimum(minimum)


class ChoiceOption(QComboBox):
    valueChanged = Signal(str)

    def __init__(self, choices=None):
        super().__init__()
        choices = {} if choices is None else choices
        for value, label in sorted(choices.items()):
            self.addItem(label, value)
        self.activated.connect(self.on_activated)

    def setValue(self, value):
        self.setCurrentIndex(self.findData(value))

    def on_activated(self, index):
        self.valueChanged.emit(self.itemData(index))


class FileOption(QGroupBox, Ui_FileWidget):
    valueChanged = Signal(str)

    def __init__(self, dialog_title, dialog_filter):
        super().__init__()
        self._dialog_title = dialog_title
        self._dialog_filter = dialog_filter
        self.setupUi(self)

    def setValue(self, value):
        self.path.setText(shorten_path(value))

    @Slot()
    def open_file_dialog(self):
        filename_suggestion = self.path.text()
        filename = QFileDialog.getSaveFileName(
            self,
            self._dialog_title,
            filename_suggestion,
            self._dialog_filter,
        )[0]
        if not filename:
            return
        self.path.setText(shorten_path(filename))
        self.valueChanged.emit(filename)

    @Slot()
    def handle_edited_path(self):
        self.valueChanged.emit(expand_path(self.path.text()))


class TableOption(QTableWidget):
    def __init__(self):
        super().__init__()
        self.horizontalHeader().setStretchLastSection(True)
        self.setSelectionMode(self.SelectionMode.SingleSelection)
        self.setTabKeyNavigation(False)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.verticalHeader().hide()
        self.currentItemChanged.connect(self._on_current_item_changed)

    def _on_current_item_changed(self, current, previous):
        # Ensure current item is visible.
        parent = self.parent()
        while parent is not None:
            if isinstance(parent, QScrollArea):
                row = current.row()
                pos = self.pos()
                x = pos.x()
                y = +pos.y() + self.rowViewportPosition(row) + self.rowHeight(row)
                parent.ensureVisible(x, y)
                return
            parent = parent.parent()


class KeymapOption(TableOption):
    valueChanged = Signal(object)

    class ItemDelegate(QStyledItemDelegate):
        def __init__(self, action_list):
            super().__init__()
            self._action_list = action_list

        def createEditor(self, parent, option, index):
            if index.column() == 1:
                combo = QComboBox(parent)
                combo.addItem("")
                combo.addItems(self._action_list)
                return combo
            return super().createEditor(parent, option, index)

    def __init__(self):
        super().__init__()
        self._value = []
        self._updating = False
        self.setColumnCount(2)
        self.setHorizontalHeaderLabels(
            (
                # i18n: Widget: “KeymapOption”.
                _("Key"),
                # i18n: Widget: “KeymapOption”.
                _("Action"),
            )
        )
        self.cellChanged.connect(self._on_cell_changed)

    def setValue(self, value):
        self._updating = True
        self._value = copy(value)
        self.setRowCount(0)
        if value is not None:
            row = -1
            for key in value.get_keys():
                action = value.get_action(key)
                if action is None:
                    action = ""
                row += 1
                self.insertRow(row)
                item = QTableWidgetItem(key)
                item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
                self.setItem(row, 0, item)
                item = QTableWidgetItem(action)
                self.setItem(row, 1, item)
        self.resizeColumnsToContents()
        self.setMinimumSize(self.viewportSizeHint())
        self.setItemDelegate(KeymapOption.ItemDelegate(value.get_actions()))
        self._updating = False

    def _on_cell_changed(self, row, column):
        if self._updating:
            return
        key = self.item(row, 0).data(Qt.ItemDataRole.DisplayRole)
        action = self.item(row, 1).data(Qt.ItemDataRole.DisplayRole)
        bindings = self._value.get_bindings()
        if action:
            bindings[key] = action
        else:
            bindings.pop(key, None)
        self._value.set_bindings(bindings)
        self.valueChanged.emit(self._value)


class MultipleChoicesOption(TableOption):
    valueChanged = Signal(object)

    LABELS = (
        # i18n: Widget: “MultipleChoicesOption”.
        _("Choice"),
        # i18n: Widget: “MultipleChoicesOption”.
        _("Selected"),
    )

    # i18n: Widget: “MultipleChoicesOption”.
    def __init__(self, choices=None, labels=None):
        super().__init__()
        if labels is None:
            labels = self.LABELS
        self._value = {}
        self._updating = False
        self._choices = {} if choices is None else choices
        self._reversed_choices = {
            translation: choice for choice, translation in choices.items()
        }
        self.setColumnCount(2)
        self.setHorizontalHeaderLabels(labels)
        self.cellChanged.connect(self._on_cell_changed)

    def setValue(self, value):
        self._updating = True
        self.resizeColumnsToContents()
        self.setMinimumSize(self.viewportSizeHint())
        self.setRowCount(0)
        if value is None:
            value = set()
        else:
            # Don't mutate the original value.
            value = set(value)
        self._value = value
        row = -1
        for choice in sorted(self._reversed_choices):
            row += 1
            self.insertRow(row)
            item = QTableWidgetItem(self._choices[choice])
            item.setFlags(item.flags() & ~Qt.ItemFlag.ItemIsEditable)
            self.setItem(row, 0, item)
            item = QTableWidgetItem()
            item.setFlags(
                (item.flags() & ~Qt.ItemFlag.ItemIsEditable)
                | Qt.ItemFlag.ItemIsUserCheckable
            )
            item.setCheckState(
                Qt.CheckState.Checked if choice in value else Qt.CheckState.Unchecked
            )
            self.setItem(row, 1, item)
        self.resizeColumnsToContents()
        self.setMinimumSize(self.viewportSizeHint())
        self._updating = False

    def _on_cell_changed(self, row, column):
        if self._updating:
            return
        assert column == 1
        choice = self._reversed_choices[
            self.item(row, 0).data(Qt.ItemDataRole.DisplayRole)
        ]
        if self.item(row, 1).checkState():
            self._value.add(choice)
        else:
            self._value.discard(choice)
        self.valueChanged.emit(self._value)


class BooleanAsDualChoiceOption(ChoiceOption):
    valueChanged = Signal(bool)

    def __init__(self, choice_false, choice_true):
        choices = {False: choice_false, True: choice_true}
        super().__init__(choices)


class ConfigOption:
    def __init__(
        self, display_name, option_name, widget_class, help_text="", dependents=()
    ):
        self.display_name = display_name
        self.option_name = option_name
        self.help_text = help_text
        self.widget_class = widget_class
        self.dependents = dependents
        self.layout = None
        self.widget = None
        self.label = None


class ConfigWindow(QDialog, Ui_ConfigWindow, WindowStateMixin):
    ROLE = "configuration"

    def __init__(self, engine):
        super().__init__()
        self.setupUi(self)
        self._engine = engine
        machines = {
            plugin.name: _(plugin.name) for plugin in registry.list_plugins("machine")
        }
        mappings = (
            # i18n: Widget: “ConfigWindow”.
            (
                _("Interface"),
                (
                    ConfigOption(
                        _("Start minimized:"),
                        "start_minimized",
                        BooleanOption,
                        _("Minimize the main window to systray on startup."),
                    ),
                    ConfigOption(
                        _("Show paper tape:"),
                        "show_stroke_display",
                        BooleanOption,
                        _("Open the paper tape on startup."),
                    ),
                    ConfigOption(
                        _("Show suggestions:"),
                        "show_suggestions_display",
                        BooleanOption,
                        _("Open the suggestions dialog on startup."),
                    ),
                    ConfigOption(
                        _("Add translation dialog opacity:"),
                        "translation_frame_opacity",
                        partial(IntOption, maximum=100, minimum=0),
                        _(
                            "Set the translation dialog opacity:\n"
                            "- 0 makes the dialog invisible\n"
                            "- 100 is fully opaque"
                        ),
                    ),
                    ConfigOption(
                        _("Dictionaries display order:"),
                        "classic_dictionaries_display_order",
                        partial(
                            BooleanAsDualChoiceOption, _("top-down"), _("bottom-up")
                        ),
                        _(
                            "Set the display order for dictionaries:\n"
                            "- top-down: match the search order; highest priority first\n"
                            "- bottom-up: reverse search order; lowest priority first\n"
                        ),
                    ),
                ),
            ),
            # i18n: Widget: “ConfigWindow”.
            (
                _("Logging"),
                (
                    ConfigOption(
                        _("Log file:"),
                        "log_file_name",
                        partial(
                            FileOption, _("Select a log file"), _("Log files (*.log)")
                        ),
                        _("File to use for logging strokes/translations."),
                    ),
                    ConfigOption(
                        _("Log strokes:"),
                        "enable_stroke_logging",
                        BooleanOption,
                        _("Save strokes to the logfile."),
                    ),
                    ConfigOption(
                        _("Log translations:"),
                        "enable_translation_logging",
                        BooleanOption,
                        _("Save translations to the logfile."),
                    ),
                ),
            ),
            # i18n: Widget: “ConfigWindow”.
            (
                _("Machine"),
                (
                    ConfigOption(
                        _("Machine:"),
                        "machine_type",
                        partial(ChoiceOption, choices=machines),
                        dependents=(
                            ("machine_specific_options", self._update_machine_options),
                            (
                                "system_keymap",
                                lambda v: self._update_keymap(machine_type=v),
                            ),
                        ),
                    ),
                    ConfigOption(
                        _("Options:"), "machine_specific_options", self._machine_option
                    ),
                    ConfigOption(_("Keymap:"), "system_keymap", KeymapOption),
                ),
            ),
            # i18n: Widget: “ConfigWindow”.
            (
                _("Output"),
                (
                    ConfigOption(
                        _("Enable at start:"),
                        "auto_start",
                        BooleanOption,
                        _("Enable output on startup."),
                    ),
                    ConfigOption(
                        _("Start attached:"),
                        "start_attached",
                        BooleanOption,
                        _(
                            "Disable preceding space on first output.\n"
                            "\n"
                            "This option is only applicable when spaces are placed before."
                        ),
                    ),
                    ConfigOption(
                        _("Start capitalized:"),
                        "start_capitalized",
                        BooleanOption,
                        _("Capitalize the first word."),
                    ),
                    ConfigOption(
                        _("Space placement:"),
                        "space_placement",
                        partial(
                            ChoiceOption,
                            choices={
                                "Before Output": _("Before Output"),
                                "After Output": _("After Output"),
                            },
                        ),
                        _("Set automatic space placement: before or after each word."),
                    ),
                    ConfigOption(
                        _("Undo levels:"),
                        "undo_levels",
                        partial(IntOption, maximum=10000, minimum=MINIMUM_UNDO_LEVELS),
                        _(
                            "Set how many preceding strokes can be undone.\n"
                            "\n"
                            "Note: the effective value will take into account the\n"
                            "dictionaries entry with the maximum number of strokes."
                        ),
                    ),
                    ConfigOption(
                        _("Key press delay (ms):"),
                        "time_between_key_presses",
                        partial(
                            IntOption,
                            maximum=100000,
                            minimum=MINIMUM_TIME_BETWEEN_KEY_PRESSES,
                        ),
                        _(
                            "Set the delay between emulated key presses (in milliseconds).\n"
                            "\n"
                            "Some programs may drop key presses if too many are sent\n"
                            "within a short period of time. Increasing the delay gives\n"
                            "programs time to process each key press.\n"
                            "Setting the delay too high will negatively impact the\n"
                            "performance of key stroke output."
                        ),
                    ),
                    ConfigOption(
                        _("Linux keyboard layout:"),
                        "keyboard_layout",
                        partial(
                            ChoiceOption,
                            choices={
                                "qwerty": "qwerty",
                                "qwertz": "qwertz",
                                "colemak": "colemak",
                                "colemak-dh": "colemak-dh",
                                "dvorak": "dvorak",
                            },
                        ),
                        _(
                            "Set the keyboard layout configurad in your system.\n"
                            "This only applies when using Linux/BSD and not using X11."
                        ),
                    ),
                ),
            ),
            # i18n: Widget: “ConfigWindow”.
            (
                _("Plugins"),
                (
                    ConfigOption(
                        _("Extension:"),
                        "enabled_extensions",
                        partial(
                            MultipleChoicesOption,
                            choices={
                                plugin.name: plugin.name
                                for plugin in registry.list_plugins("extension")
                            },
                            labels=(_("Name"), _("Enabled")),
                        ),
                        _("Configure enabled plugin extensions."),
                    ),
                ),
            ),
            # i18n: Widget: “ConfigWindow”.
            (
                _("System"),
                (
                    ConfigOption(
                        _("System:"),
                        "system_name",
                        partial(
                            ChoiceOption,
                            choices={
                                plugin.name: plugin.name
                                for plugin in registry.list_plugins("system")
                            },
                        ),
                        dependents=(
                            (
                                "system_keymap",
                                lambda v: self._update_keymap(system_name=v),
                            ),
                        ),
                    ),
                ),
            ),
        )
        # Only keep supported options, to avoid messing with things like
        # dictionaries, that are handled by another (possibly concurrent)
        # dialog.
        self._supported_options = set()
        for section, option_list in mappings:
            self._supported_options.update(option.option_name for option in option_list)
        self._update_config()
        # Create and fill tabs.
        option_by_name = {}
        for section, option_list in mappings:
            layout = QFormLayout()
            for option in option_list:
                option_by_name[option.option_name] = option
                option.layout = layout
                option.widget = self._create_option_widget(option)
                option.label = QLabel(option.display_name)
                option.label.setToolTip(option.help_text)
                option.label.setBuddy(option.widget)
                layout.addRow(option.label, option.widget)
            frame = QFrame()
            frame.setLayout(layout)
            frame.setAccessibleName(section)
            frame.setFocusProxy(option_list[0].widget)
            scroll_area = QScrollArea()
            scroll_area.setWidgetResizable(True)
            scroll_area.setWidget(frame)
            scroll_area.setFocusProxy(frame)
            self.tabs.addTab(scroll_area, section)
        # Update dependents.
        for option in option_by_name.values():
            option.dependents = [
                (option_by_name[option_name], update_fn)
                for option_name, update_fn in option.dependents
            ]
        self.buttons.button(QDialogButtonBox.StandardButton.Ok).clicked.connect(
            self.on_apply
        )
        self.buttons.button(QDialogButtonBox.StandardButton.Apply).clicked.connect(
            self.on_apply
        )
        self.tabs.currentWidget().setFocus()
        self.restore_state()
        self.finished.connect(self.save_state)

    def _update_config(self, save=False):
        with self._engine:
            if save:
                self._engine.config = self._config.maps[0]
            self._config = ChainMap(
                {},
                {
                    name: value
                    for name, value in self._engine.config.items()
                    if name in self._supported_options
                },
            )

    def _machine_option(self, *args):
        machine_options = {
            plugin.name: plugin.obj
            for plugin in registry.list_plugins("gui.qt.machine_option")
        }
        machine_type = self._config["machine_type"]
        machine_class = registry.get_plugin("machine", machine_type).obj
        for klass in machine_class.mro():
            # Look for `module_name:class_name` before `class_name`.
            for name in (
                "%s:%s" % (klass.__module__, klass.__name__),
                klass.__name__,
            ):
                opt_class = machine_options.get(name)
                if opt_class is not None:
                    return opt_class(*args)
        return NopeOption(*args)

    def _update_machine_options(self, machine_type=None):
        return self._engine[
            ("machine_specific_options", machine_type or self._config["machine_type"])
        ]

    def _update_keymap(self, system_name=None, machine_type=None):
        return self._engine[
            (
                "system_keymap",
                system_name or self._config["system_name"],
                machine_type or self._config["machine_type"],
            )
        ]

    def _create_option_widget(self, option):
        widget = option.widget_class()
        widget.setToolTip(option.help_text)
        widget.setAccessibleName(option.display_name)
        widget.setAccessibleDescription(option.help_text)
        widget.valueChanged.connect(partial(self.on_option_changed, option))
        widget.setValue(self._config[option.option_name])
        return widget

    def keyPressEvent(self, event):
        # Disable Enter/Return key to trigger "OK".
        if event.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            return
        super().keyPressEvent(event)

    def on_option_changed(self, option, value):
        self._config[option.option_name] = value
        for dependent, update_fn in option.dependents:
            # Ensure unsaved changes are discarded.
            if dependent.option_name in self._config.maps[0]:
                del self._config.maps[0][dependent.option_name]
            self._config.maps[1][dependent.option_name] = update_fn(value)
            widget = self._create_option_widget(dependent)
            dependent.layout.replaceWidget(dependent.widget, widget)
            dependent.label.setBuddy(widget)
            dependent.widget.deleteLater()
            dependent.widget = widget

    def on_apply(self):
        self._update_config(save=True)



from plover import _, log
from plover.misc import shorten_path
from plover.steno import normalize_steno


class StenoNormalizer:
    def __init__(self, dictionary_path):
        self._dictionary_path = dictionary_path
        self._errors_count = 0

    def normalize(self, steno):
        try:
            return normalize_steno(steno)
        except ValueError:
            self._errors_count += 1
            return tuple(steno.split("/"))

    def __enter__(self):
        return self.normalize

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None and self._errors_count:
            log.warning(
                _("dictionary `%s` loaded with %u invalid steno errors"),
                shorten_path(self._dictionary_path),
                self._errors_count,
            )



from operator import attrgetter, itemgetter
from collections import namedtuple
from itertools import chain
from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt, Slot
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QStyledItemDelegate,
)

from plover import _
from plover.translation import escape_translation, unescape_translation
from plover.misc import expand_path, shorten_path
from plover.steno import normalize_steno, steno_to_sort_key

from plover.gui_qt.dictionary_editor_ui import Ui_DictionaryEditor
from plover.gui_qt.steno_validator import StenoValidator
from plover.gui_qt.utils import ToolBar, WindowStateMixin


_COL_STENO, _COL_TRANS, _COL_DICT, _COL_COUNT = range(3 + 1)


class DictionaryItem(namedtuple("DictionaryItem", "steno translation dictionary")):
    @property
    def strokes(self):
        return normalize_steno(self.steno, strict=False)

    @property
    def dictionary_path(self):
        return self.dictionary.path


class DictionaryItemDelegate(QStyledItemDelegate):
    def __init__(self, dictionary_list):
        super().__init__()
        self._dictionary_list = dictionary_list

    def createEditor(self, parent, option, index):
        if index.column() == _COL_DICT:
            dictionary_paths = [
                shorten_path(dictionary.path)
                for dictionary in self._dictionary_list
                if not dictionary.readonly
            ]
            combo = QComboBox(parent)
            combo.addItems(dictionary_paths)
            return combo
        widget = super().createEditor(parent, option, index)
        if index.column() == _COL_STENO:
            widget.setValidator(StenoValidator())
        return widget


class DictionaryItemModel(QAbstractTableModel):
    def __init__(self, dictionary_list, sort_column, sort_order):
        super().__init__()
        self._error_icon = QIcon(":resources/dictionary_error.svg")
        self._dictionary_list = dictionary_list
        self._operations = []
        self._entries = []
        self._sort_column = sort_column
        self._sort_order = sort_order
        self._update_entries()

    def _update_entries(self, strokes_filter=None, translation_filter=None):
        self._entries = []
        for dictionary in self._dictionary_list:
            for strokes, translation in dictionary.items():
                steno = "/".join(strokes)
                if strokes_filter is not None and not steno.startswith(strokes_filter):
                    continue
                if translation_filter is not None and not translation.startswith(
                    translation_filter
                ):
                    continue
                item = DictionaryItem(steno, translation, dictionary)
                self._entries.append(item)
        self.sort(self._sort_column, self._sort_order)

    @property
    def has_undo(self):
        return bool(self._operations)

    @property
    def modified(self):
        paths = set()
        dictionary_list = []
        for op_list in self._operations:
            if not isinstance(op_list, list):
                op_list = (op_list,)
            for item in chain(*op_list):
                if item is None:
                    continue
                dictionary = item.dictionary
                if dictionary.path in paths:
                    continue
                paths.add(dictionary.path)
                dictionary_list.append(dictionary)
        return dictionary_list

    # Note:
    # - since switching from a dictionary to a table does not enforce the
    #   unicity of keys, a deletion can fail when one of the duplicate has
    #   already been deleted.
    # - when undoing an operation at the table level, the item may have
    #   been filtered-out and not present

    def _undo(self, old_item, new_item):
        if old_item is None:
            # Undo addition.
            try:
                del new_item.dictionary[new_item.strokes]
            except KeyError:
                pass
            try:
                row = self._entries.index(new_item)
            except ValueError:
                # Happen if the item is filtered-out.
                pass
            else:
                self.remove_rows([row], record=False)
            return
        if new_item is None:
            # Undo deletion.
            self.new_row(0, item=old_item, record=False)
            return
        # Undo update.
        try:
            del new_item.dictionary[new_item.strokes]
        except KeyError:
            pass
        try:
            row = self._entries.index(new_item)
        except ValueError:
            # Happen if the item is filtered-out,
            # "create" a new row so the user see
            # the result of the undo.
            self.new_row(0, item=old_item, record=False)
        else:
            old_item.dictionary[old_item.strokes] = old_item.translation
            self._entries[row] = old_item
            self.dataChanged.emit(
                self.index(row, _COL_STENO), self.index(row, _COL_TRANS)
            )

    def undo(self, op=None):
        op = self._operations.pop()
        if isinstance(op, list):
            for old_item, new_item in op:
                self._undo(old_item, new_item)
        else:
            self._undo(*op)

    def rowCount(self, parent):
        return 0 if parent.isValid() else len(self._entries)

    def columnCount(self, parent):
        return _COL_COUNT

    def headerData(self, section, orientation, role):
        if (
            orientation != Qt.Orientation.Horizontal
            or role != Qt.ItemDataRole.DisplayRole
        ):
            return None
        if section == _COL_STENO:
            # i18n: Widget: “DictionaryEditor”.
            return _("Strokes")
        if section == _COL_TRANS:
            # i18n: Widget: “DictionaryEditor”.
            return _("Translation")
        if section == _COL_DICT:
            # i18n: Widget: “DictionaryEditor”.
            return _("Dictionary")

    def data(self, index, role):
        if not index.isValid() or role not in (
            Qt.ItemDataRole.EditRole,
            Qt.ItemDataRole.DisplayRole,
            Qt.ItemDataRole.DecorationRole,
        ):
            return None
        item = self._entries[index.row()]
        column = index.column()
        if role == Qt.ItemDataRole.DecorationRole:
            if column == _COL_STENO:
                try:
                    normalize_steno(item.steno)
                except ValueError:
                    return self._error_icon
            return None
        if column == _COL_STENO:
            return item.steno
        if column == _COL_TRANS:
            return escape_translation(item.translation)
        if column == _COL_DICT:
            return shorten_path(item.dictionary.path)

    def flags(self, index):
        if not index.isValid():
            return Qt.NoItemFlags
        f = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable
        item = self._entries[index.row()]
        if not item.dictionary.readonly:
            f |= Qt.ItemFlag.ItemIsEditable
        return f

    def filter(self, strokes_filter=None, translation_filter=None):
        self.modelAboutToBeReset.emit()
        self._update_entries(strokes_filter, translation_filter)
        self.modelReset.emit()

    @staticmethod
    def _item_steno_sort_key(item):
        return steno_to_sort_key(item[_COL_STENO], strict=False)

    def sort(self, column, order):
        self.layoutAboutToBeChanged.emit()
        if column == _COL_DICT:
            key = attrgetter("dictionary_path")
        elif column == _COL_STENO:
            key = self._item_steno_sort_key
        else:
            key = itemgetter(column)
        self._entries.sort(key=key, reverse=(order == Qt.SortOrder.DescendingOrder))
        self._sort_column = column
        self._sort_order = order
        self.layoutChanged.emit()

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole, record=True):
        assert role == Qt.ItemDataRole.EditRole
        row = index.row()
        column = index.column()
        old_item = self._entries[row]
        strokes = old_item.strokes
        steno, translation, dictionary = old_item
        if column == _COL_STENO:
            strokes = normalize_steno(value.strip(), strict=False)
            steno = "/".join(strokes)
            if not steno or steno == old_item.steno:
                return False
        elif column == _COL_TRANS:
            translation = unescape_translation(value.strip())
            if translation == old_item.translation:
                return False
        elif column == _COL_DICT:
            path = expand_path(value)
            for dictionary in self._dictionary_list:
                if dictionary.path == path:
                    break
            if dictionary == old_item.dictionary:
                return False
        try:
            del old_item.dictionary[old_item.strokes]
        except KeyError:
            pass
        if not old_item.strokes and not old_item.translation:
            # Merge operations when editing a newly added row.
            if self._operations and self._operations[-1] == [(None, old_item)]:
                self._operations.pop()
                old_item = None
        new_item = DictionaryItem(steno, translation, dictionary)
        self._entries[row] = new_item
        dictionary[strokes] = translation
        if record:
            self._operations.append((old_item, new_item))
        self.dataChanged.emit(index, index)
        return True

    def new_row(self, row, item=None, record=True):
        if item is None:
            if row == 0 and not self._entries:
                dictionary = self._dictionary_list[0]
            else:
                dictionary = self._entries[row].dictionary
            item = DictionaryItem("", "", dictionary)
        self.beginInsertRows(QModelIndex(), row, row)
        self._entries.insert(row, item)
        if record:
            self._operations.append((None, item))
        self.endInsertRows()

    def remove_rows(self, row_list, record=True):
        assert row_list
        operations = []
        for row in sorted(row_list, reverse=True):
            self.beginRemoveRows(QModelIndex(), row, row)
            item = self._entries.pop(row)
            self.endRemoveRows()
            try:
                del item.dictionary[item.strokes]
            except KeyError:
                pass
            else:
                operations.append((item, None))
        if record:
            self._operations.append(operations)


class DictionaryEditor(QDialog, Ui_DictionaryEditor, WindowStateMixin):
    ROLE = "dictionary_editor"

    def __init__(self, engine, dictionary_paths):
        super().__init__()
        self.setupUi(self)
        self._engine = engine
        with engine:
            dictionary_list = [
                dictionary
                for dictionary in engine.dictionaries.dicts
                if dictionary.path in dictionary_paths
            ]
        sort_column, sort_order = _COL_STENO, Qt.SortOrder.AscendingOrder
        self._model = DictionaryItemModel(dictionary_list, sort_column, sort_order)
        self._model.dataChanged.connect(self.update_after_data_change)
        self.table.sortByColumn(sort_column, sort_order)
        self.table.setSortingEnabled(True)
        self.table.setModel(self._model)
        self.table.resizeColumnsToContents()
        self.table.setItemDelegate(DictionaryItemDelegate(dictionary_list))
        self.table.selectionModel().selectionChanged.connect(
            self.handle_selection_change
        )
        background = self.table.palette().highlightedText().color().name()
        text_color = self.table.palette().highlight().color().name()
        self.table.setStyleSheet(
            """
                                 QTableView::item:focus {
                                     background-color: %s;
                                     color: %s;
                                }"""
            % (background, text_color)
        )
        self.table.setFocus()
        for action in (
            self.action_Undo,
            self.action_Delete,
        ):
            action.setEnabled(False)
        # Toolbar.
        self.layout().addWidget(
            ToolBar(
                self.action_Undo,
                self.action_Delete,
                self.action_New,
            )
        )
        self.strokes_filter.setValidator(StenoValidator())
        self.restore_state()
        self.finished.connect(self.save_state)

    @property
    def _selection(self):
        return list(
            sorted(index.row() for index in self.table.selectionModel().selectedRows(0))
        )

    def _select(self, row, edit=False):
        row = min(row, self._model.rowCount(QModelIndex()) - 1)
        index = self._model.index(row, 0)
        self.table.setCurrentIndex(index)
        if edit:
            self.table.edit(index)

    @Slot(QModelIndex, QModelIndex)
    def update_after_data_change(self, top_left, bottom_right):
        self.table.setCurrentIndex(top_left)
        self.action_Undo.setEnabled(self._model.has_undo)

    @Slot()
    def handle_selection_change(self):
        enabled = bool(self._selection)
        for action in (self.action_Delete,):
            action.setEnabled(enabled)

    @Slot()
    def undo(self):
        assert self._model.has_undo
        self._model.undo()
        self.action_Undo.setEnabled(self._model.has_undo)

    @Slot()
    def delete_selected_row(self):
        selection = self._selection
        assert selection
        self._model.remove_rows(selection)
        self._select(selection[0])
        self.action_Undo.setEnabled(self._model.has_undo)

    @Slot()
    def add_new_row(self):
        selection = self._selection
        if selection:
            row = self._selection[0]
        else:
            row = 0
        self.table.reset()
        self._model.new_row(row)
        self._select(row, edit=True)
        self.action_Undo.setEnabled(self._model.has_undo)

    @Slot()
    def apply_filter(self):
        self.table.selectionModel().clear()
        strokes_filter = "/".join(normalize_steno(self.strokes_filter.text().strip()))
        translation_filter = unescape_translation(
            self.translation_filter.text().strip()
        )
        self._model.filter(
            strokes_filter=strokes_filter, translation_filter=translation_filter
        )

    @Slot()
    def clear_filter(self):
        self.strokes_filter.setText("")
        self.translation_filter.setText("")
        self._model.filter(strokes_filter=None, translation_filter=None)

    @Slot(int)
    def save_modified_dictionaries(self, result):
        with self._engine:
            self._engine.dictionaries.save(
                dictionary.path for dictionary in self._model.modified
            )



import re

from PySide6.QtWidgets import QDialog

import plover

from plover.gui_qt.about_dialog_ui import Ui_AboutDialog


class AboutDialog(QDialog, Ui_AboutDialog):
    ROLE = "about"

    def __init__(self, engine):
        super().__init__()
        self.setupUi(self)
        credits = plover.__credits__
        credits = re.sub(r"<([^>]*)>", r'<a href="\1">\1</a>', credits)
        credits = credits.replace("\n", "<br/>")
        self.text.setHtml(
            """
            <style>
            h1 {text-align:center;}
            h2 {text-align:center;}
            p {text-align:center;}
            </style>
            <p><img src="%(icon)s"/></p>
            <h1>%(name)s %(version)s</h1>
            <p>%(description)s</p>
            <p><i>Copyright %(copyright)s</i></p>
            <p>License: <a href="%(license_url)s">%(license)s</a></p>
            <p>Project Homepage: <a href='%(url)s'>%(url)s</a></p>
            <h2>Credits:</h2>
            <p>%(credits)s</p>
            """
            % {
                "icon": ":/resources/plover.png",
                "name": plover.__name__.capitalize(),
                "version": plover.__version__,
                "description": plover.__long_description__,
                "copyright": plover.__copyright__.replace("(C)", "&copy;"),
                "license": plover.__license__,
                "license_url": "https://www.gnu.org/licenses/gpl-2.0-standalone.html",
                "url": plover.__download_url__,
                "credits": credits,
            }
        )



import re

from plover.formatting import _LookAheadAction


IF_NEXT_META_RX = re.compile(r"((?:[^\\/]|\\\\|\\/)*)/?")
IF_NEXT_ESCAPE_RX = re.compile(r"\\([\\/])")


def meta_if_next_matches(ctx, meta):
    pattern, result1, result2 = [
        IF_NEXT_ESCAPE_RX.sub(r"\1", s)
        for s in filter(None, IF_NEXT_META_RX.split(meta, 2))
    ]
    action_list = []
    for alternative in result1, result2:
        action = ctx.new_action()
        action.text = alternative
        action_list.append(action)
    return _LookAheadAction(pattern, *action_list)



# Copyright (c) 2011 Hesky Fisher
# See LICENSE.txt for details.
# Many thanks to a steno geek for help with the protocol.

# TODO: Come up with a mechanism to communicate back to the engine when there
# is a connection error.
# TODO: Address any generic exceptions still left.

"""Thread-based monitoring of a stenotype machine using the stentura protocol."""

"""
The stentura protocol uses packets to communicate with the machine. A
request packet is sent to the machine and a response packet is received. If
no response is received after a one second timeout then the same packet
should be sent again. The machine may hold off on responding to a READC
packet for up to 500ms if there are no new strokes.

Each request packet should have a sequence number that is one higher than
the previously sent packet modulo 256. The response packet will have the
same sequence number. Each packet consists of a header followed by an
optional data section. All multibyte fields are little endian.

The request packet header is structured as follows:
- SOH: 1 byte. Always set to ASCII SOH (0x1).
- seq: 1 byte. The sequence number of this packet.
- length: 2 bytes. The total length of the packet, including the data
  section, in bytes.
- action: 2 bytes. The action requested. See actions below.
- p1: 2 bytes. Parameter 1. The values for the parameters depend on the
  action.
- p2: 2 bytes. Parameter 2.
- p3: 2 bytes. Parameter 3.
- p4: 2 bytes. Parameter 4.
- p5: 2 bytes. Parameter 5.
- checksum: 2 bytes. The CRC is computed over the packet from seq through
  p5. The specific CRC algorithm used is described above in the Crc class.

The request header can be followed by a data section. The meaning of the
data section depends on the action:
- data: variable length.
- crc: 2 bytes. A CRC over just the data section.

The response packet header is structured as follows:
- SOH: 1 byte. Always set to ASCII SOH (0x1).
- seq: 1 byte. The sequence number of the request packet.
- length: 2 bytes. The total length of the packet, including the data
  section, in bytes.
- action: 2 bytes. The action of the request packet.
- error: 2 bytes. The error code. Zero if no error.
- p1: 2 bytes. Parameter 1. The values of the parameters depend on the
  action.
- p2: 2 bytes. Parameter 2.
- checksum: 2 bytes. The CRC is computed over the packet from seq through
  p2.

The response header can be follows by a data section, whose meaning is
dependent on the action. The structure is the same as in request packets.

The stentura machine has a concept of drives and files. The first (and
possibly only) drive is called A. Each file consists of a set of one or
more blocks. Each block is 512 bytes long.

In addition to regular files, there is a realtime file whose name is
'REALTIME.000'. All strokes typed are appended to this file. Subsequent
reads from the realtime file ignore positional arguments and only return
all the strokes since the last read action. However, opening the file again
and reading from the beginning will result in all the same strokes being
read again. The only reliable way to jump to the end is to do a full,
sequential, read to the end before processing any strokes. I'm told that on
some machines sending a READC without an OPEN will just read from the
realtime file.

The contents of the files are a sequence of strokes. Each stroke consists
of four bytes. Each byte has the two most significant bytes set to one. The
rest of the byte is a bitmask indicating which keys were pressed during the
stroke. The format is as follows: 11^#STKP 11WHRAO* 11EUFRPB 11LGTSDZ ^ is
something called a stenomark. I'm not sure what that is. # is the number
bar.

Note: Only OPEN and READC are needed to get strokes as they are typed from
the realtime file.

Actions and their packets:

All unmentioned parameters should be zero and unless explicitly mentioned
the packet should have no data section.

RESET (0x14):
Unknown.

DISKSTATUS (0x7):
Unknown.
p1 is set to the ASCII value corresponding to the drive letter, e.g. 'A'.

GETDOS (0x18):
Returns the DOS filenames for the files in the requested drive.
p1 is set to the ASCII value corresponding to the drive letter, e.g. 'A'.
p2 is set to one to return the name of the realtime file (which is always
'REALTIME.000').
p3 controls which page to return, with 20 filenames per page.
The return packet contains a data section that is 512 bytes long. The first
bytes seems to be one. The filename for the first file starts at offset 32.
My guess would be that the other filenames would exist at a fixed offset of
24 bytes apart. So first filename is at 32, second is at 56, third at 80,
etc. There seems to be some meta data stored after the filename but I don't
know what it means.

DELETE (0x3):
Deletes the specified files. NOP on realtime file.
p1 is set to the ASCII value corresponding to the drive letter, e.g. 'A'.
The filename is specified in the data section.

OPEN (0xA):
Opens a file for reading. This action is sticky and causes this file to be
the current file for all following READC packets.
p1 is set to the ASCII value corresponding to the drive letter, e.g. 'A'.
The filename is specified in the data section.
I'm told that if there is an error opening the realtime file then no
strokes have been written yet.
TODO: Check that and implement workaround.

READC (0xB):
Reads characters from the currently opened file.
p1 is set to 1, I'm not sure why.
p3 is set to the maximum number of bytes to read but should probably be
512.
p4 is set to the block number.
p5 is set to the starting byte offset within the block.
It's possible that the machine will ignore the positional arguments to
READC when reading from the realtime file and just return successive values
for each call.
The response will have the number of bytes read in p1 (but the same is
deducible from the length). The data section will have the contents read
from the file.

CLOSE (0x2):
Closes the current file.
p1 is set to one, I don't know why.

TERM (0x15):
Unknown.

DIAG (0x19):
Unknown.

"""

import struct

from plover import log
import plover.machine.base


# Python 3 replacement for Python 2 buffer.
def buffer(object, offset=None, size=None):
    if offset is None:
        offset = 0
    if size is None:
        size = len(object) - offset
    return memoryview(object)[offset : offset + size]


def _allocate_buffer():
    return bytearray(1024)


class _ProtocolViolationException(Exception):
    """Something has happened that is doesn't follow the protocol."""

    pass


class _StopException(Exception):
    """The thread was asked to stop."""

    pass


class _TimeoutException(Exception):
    """An operation has timed out."""

    pass


class _ConnectionLostException(Exception):
    """Cannot communicate with the machine."""

    pass


# fmt: off
_CRC_TABLE = [
    0x0000, 0xc0c1, 0xc181, 0x0140, 0xc301, 0x03c0, 0x0280, 0xc241,
    0xc601, 0x06c0, 0x0780, 0xc741, 0x0500, 0xc5c1, 0xc481, 0x0440,
    0xcc01, 0x0cc0, 0x0d80, 0xcd41, 0x0f00, 0xcfc1, 0xce81, 0x0e40,
    0x0a00, 0xcac1, 0xcb81, 0x0b40, 0xc901, 0x09c0, 0x0880, 0xc841,
    0xd801, 0x18c0, 0x1980, 0xd941, 0x1b00, 0xdbc1, 0xda81, 0x1a40,
    0x1e00, 0xdec1, 0xdf81, 0x1f40, 0xdd01, 0x1dc0, 0x1c80, 0xdc41,
    0x1400, 0xd4c1, 0xd581, 0x1540, 0xd701, 0x17c0, 0x1680, 0xd641,
    0xd201, 0x12c0, 0x1380, 0xd341, 0x1100, 0xd1c1, 0xd081, 0x1040,
    0xf001, 0x30c0, 0x3180, 0xf141, 0x3300, 0xf3c1, 0xf281, 0x3240,
    0x3600, 0xf6c1, 0xf781, 0x3740, 0xf501, 0x35c0, 0x3480, 0xf441,
    0x3c00, 0xfcc1, 0xfd81, 0x3d40, 0xff01, 0x3fc0, 0x3e80, 0xfe41,
    0xfa01, 0x3ac0, 0x3b80, 0xfb41, 0x3900, 0xf9c1, 0xf881, 0x3840,
    0x2800, 0xe8c1, 0xe981, 0x2940, 0xeb01, 0x2bc0, 0x2a80, 0xea41,
    0xee01, 0x2ec0, 0x2f80, 0xef41, 0x2d00, 0xedc1, 0xec81, 0x2c40,
    0xe401, 0x24c0, 0x2580, 0xe541, 0x2700, 0xe7c1, 0xe681, 0x2640,
    0x2200, 0xe2c1, 0xe381, 0x2340, 0xe101, 0x21c0, 0x2080, 0xe041,
    0xa001, 0x60c0, 0x6180, 0xa141, 0x6300, 0xa3c1, 0xa281, 0x6240,
    0x6600, 0xa6c1, 0xa781, 0x6740, 0xa501, 0x65c0, 0x6480, 0xa441,
    0x6c00, 0xacc1, 0xad81, 0x6d40, 0xaf01, 0x6fc0, 0x6e80, 0xae41,
    0xaa01, 0x6ac0, 0x6b80, 0xab41, 0x6900, 0xa9c1, 0xa881, 0x6840,
    0x7800, 0xb8c1, 0xb981, 0x7940, 0xbb01, 0x7bc0, 0x7a80, 0xba41,
    0xbe01, 0x7ec0, 0x7f80, 0xbf41, 0x7d00, 0xbdc1, 0xbc81, 0x7c40,
    0xb401, 0x74c0, 0x7580, 0xb541, 0x7700, 0xb7c1, 0xb681, 0x7640,
    0x7200, 0xb2c1, 0xb381, 0x7340, 0xb101, 0x71c0, 0x7080, 0xb041,
    0x5000, 0x90c1, 0x9181, 0x5140, 0x9301, 0x53c0, 0x5280, 0x9241,
    0x9601, 0x56c0, 0x5780, 0x9741, 0x5500, 0x95c1, 0x9481, 0x5440,
    0x9c01, 0x5cc0, 0x5d80, 0x9d41, 0x5f00, 0x9fc1, 0x9e81, 0x5e40,
    0x5a00, 0x9ac1, 0x9b81, 0x5b40, 0x9901, 0x59c0, 0x5880, 0x9841,
    0x8801, 0x48c0, 0x4980, 0x8941, 0x4b00, 0x8bc1, 0x8a81, 0x4a40,
    0x4e00, 0x8ec1, 0x8f81, 0x4f40, 0x8d01, 0x4dc0, 0x4c80, 0x8c41,
    0x4400, 0x84c1, 0x8581, 0x4540, 0x8701, 0x47c0, 0x4680, 0x8641,
    0x8201, 0x42c0, 0x4380, 0x8341, 0x4100, 0x81c1, 0x8081, 0x4040
]
# fmt: on


def _crc(data, offset=None, size=None):
    """Compute the Crc algorithm used by the stentura protocol.

    This algorithm is described by the Rocksoft^TM Model CRC Algorithm as
    follows:

    Name   : "CRC-16"
    Width  : 16
    Poly   : 8005
    Init   : 0000
    RefIn  : True
    RefOut : True
    XorOut : 0000
    Check  : BB3D

    Args:
    - data: The data to checksum. The data should be an iterable that returns
            bytes

    Returns: The computed crc for the data.

    """
    if offset is None:
        offset = 0
    if size is None:
        size = len(data) - offset
    checksum = 0
    for n in range(offset, offset + size):
        b = data[n]
        checksum = _CRC_TABLE[(checksum ^ b) & 0xFF] ^ ((checksum >> 8) & 0xFF)
    return checksum


def _write_to_buffer(buf, offset, data):
    """Write data to buf at offset.

    Note: buf must be big enough, and will not be extended as needed.

    Args:
    - buf: The buffer. Should be of type bytearray()
    - offset. The offset at which to start writing.
    - data: An iterable containing the data to write.
    """
    buf[offset : offset + len(data)] = data


# Helper table for parsing strokes of the form:
# 11^#STKP 11WHRAO* 11EUFRPB 11LGTSDZ
# fmt: off
_STENO_KEY_CHART = ("^", "#", "S-", "T-", "K-", "P-",    # Byte #1
                    "W-", "H-", "R-", "A-", "O-", "*",   # Byte #2
                    "-E", "-U", "-F", "-R", "-P", "-B",  # Byte #3
                    "-L", "-G", "-T", "-S", "-D", "-Z")  # Byte #4
# fmt: on


def _parse_stroke(a, b, c, d):
    """Parse a stroke and return a list of keys pressed.

    Args:
    - a: The first byte.
    - b: The second byte.
    - c: The third byte.
    - d: The fourth byte.

    Returns: A sequence with all the keys pressed in the stroke.
             e.g. ['S-', 'A-', '-T']

    """
    fullstroke = ((a & 0x3F) << 18) | ((b & 0x3F) << 12) | ((c & 0x3F) << 6) | d & 0x3F
    return [_STENO_KEY_CHART[i] for i in range(24) if (fullstroke & (1 << (23 - i)))]


def _parse_strokes(data):
    """Parse strokes from a buffer and return a sequence of strokes.

    Args:
    - data: A byte buffer.

    Returns: A sequence of strokes. Each stroke is a sequence of pressed keys.

    Throws:
    - _ProtocolViolationException if the data doesn't follow the protocol.

    """
    strokes = []
    if (len(data) % 4) != 0:
        raise _ProtocolViolationException(
            "Data size is not divisible by 4: %d" % (len(data))
        )
    for b in data:
        if (b & 0b11000000) != 0b11000000:
            raise _ProtocolViolationException("Data is not stroke: 0x%X" % (b))
    for a, b, c, d in zip(*([iter(data)] * 4)):
        strokes.append(_parse_stroke(a, b, c, d))
    return strokes


# Actions
_CLOSE = 0x2
_DELETE = 0x3
_DIAG = 0x19
_DISKSTATUS = 0x7
_GETDOS = 0x18
_OPEN = 0xA
_READC = 0xB
_RESET = 0x14
_TERM = 0x15

# Compiled struct for writing request headers.
_REQUEST_STRUCT = struct.Struct("<2B7H")
_SHORT_STRUCT = struct.Struct("<H")


def _make_request(buf, action, seq, p1=0, p2=0, p3=0, p4=0, p5=0, data=None):
    """Create a request packet.

    Args:
    - buf: The buffer used for the packet. Should be bytearray() and big
    enough, as it will not be extended as needed.
    - action: The action for the packet.
    - seq: The sequence numbe for the packet.
    - p1 - p5: Parameter N for the packet (default: 0).
    - data: The data to add to the packet as a sequence of bytes, if any
    (default: None).

    Returns: A buffer as a slice of the passed in buf that holds the packet.

    """
    length = 18
    if data:
        length += len(data) + 2  # +2 for the data CRC.
    _REQUEST_STRUCT.pack_into(buf, 0, 1, seq, length, action, p1, p2, p3, p4, p5)
    crc = _crc(buf, 1, 15)
    _SHORT_STRUCT.pack_into(buf, 16, crc)
    if data:
        _write_to_buffer(buf, 18, data)
        crc = _crc(data)
        _SHORT_STRUCT.pack_into(buf, length - 2, crc)
    return buffer(buf, 0, length)


def _make_open(buf, seq, drive, filename):
    """Make a packet with the OPEN command.

    Args:
    - buf: The buffer to use of type bytearray(). Will be extended if
    needed.
    - seq: The sequence number of the packet.
    - drive: The letter of the drive (probably 'A').
    - filename: The name of the file (probably 'REALTIME.000').

    Returns: A buffer as a slice of the passed in buf that holds the packet.

    """
    return _make_request(buf, _OPEN, seq, p1=ord(drive), data=filename)


def _make_read(buf, seq, block, byte, length=512):
    """Make a packet with the READC command.

    Args:
    - buf: The buffer to use of type bytearray(). Will be extended if
    needed.
    - seq: The sequence number of the packet.
    - block: The index of the file block to read.
    - byte: The byte offset within the block at which to start reading.
    - length: The number of bytes to read, max 512 (default: 512).

    Returns: A buffer as a slice of the passed in buf that holds the packet.

    """
    return _make_request(buf, _READC, seq, p1=1, p3=length, p4=block, p5=byte)


def _make_reset(buf, seq):
    """Make a packet with the RESET command.

    Args:
    - buf: The buffer to use of type bytearray(). Will be extended if
    needed.
    - seq: The sequence number of the packet.

    Returns: A buffer as a slice of the passed in buf that holds the packet.

    """
    return _make_request(buf, _RESET, seq)


def _validate_response(packet):
    """Validate a response packet.

    Args:
    - packet: The packet to validate.

    Returns: True if the packet is valid, False otherwise.

    """
    if len(packet) < 14:
        return False
    length = _SHORT_STRUCT.unpack(packet[2:4])[0]
    if length != len(packet):
        return False
    if _crc(packet, 1, 13) != 0:
        return False
    if length > 14:
        if length < 17:
            return False
        if _crc(packet, 14) != 0:
            return False
    return True


def _read_data(port, stop, buf, offset, num_bytes):
    """Read data off the serial port and into port at offset.

    Args:
    - port: The serial port to read.
    - stop: An event which, when set, causes this function to stop.
    - buf: The buffer to write.
    - offset: The offset into the buffer to write.
    - num_bytes: The number of bytes expected

    Returns: The number of bytes read.

    Raises:
    _StopException: If stop is set.
    _TimeoutException: If the timeout is reached with no data read.

    """

    assert num_bytes > 0
    read_bytes = port.read(num_bytes)
    if stop.is_set():
        raise _StopException()
    if num_bytes > len(read_bytes):
        raise _TimeoutException()
    _write_to_buffer(buf, offset, read_bytes)
    return len(read_bytes)


MINIMUM_PACKET_LENGTH = 14


def _read_packet(port, stop, buf):
    """Read a full packet from the port.

    Reads from the port until a full packet is received or the stop or timeout
    conditions are met.

    Args:
    - port: The port to read.
    - stop: Event object used to request stopping.
    - buf: The buffer to write.

    Returns: A buffer as a slice of buf holding the packet.

    Raises:
    _ProtocolViolationException: If the packet doesn't conform to the protocol.
    _TimeoutException: If the packet is not read within the timeout.
    _StopException: If a stop was requested.

    """
    bytes_read = 0
    bytes_read += _read_data(port, stop, buf, bytes_read, 4)
    assert 4 == bytes_read
    packet_length = _SHORT_STRUCT.unpack_from(buf, 2)[0]
    # Packet length should always be at least 14 bytes long
    if packet_length < MINIMUM_PACKET_LENGTH:
        raise _ProtocolViolationException()
    bytes_read += _read_data(port, stop, buf, bytes_read, packet_length - bytes_read)
    packet = buffer(buf, 0, bytes_read)
    if not _validate_response(packet):
        raise _ProtocolViolationException()
    return buffer(buf, 0, bytes_read)


def _write_to_port(port, data):
    """Write data to a port.

    Args:
    - port: The port to write.
    - data: The data to write

    """
    while data:
        data = buffer(data, port.write(data))


def _send_receive(port, stop, packet, buf, max_tries=3):
    """Send a packet and return the response.

    Send a packet and make sure there is a response and it is for the correct
    request and return it, otherwise retry max_retries times.

    Args:
    - port: The port to read.
    - stop: Event used to signal tp stop.
    - packet: The packet to send. May be used after buf is written so should be
    distinct.
    - buf: Buffer used to store response.
    - max_tries: The maximum number of times to retry sending the packet and
    reading the response before giving up (default: 3).

    Returns: A buffer as a slice of buf holding the response packet.

    Raises:
    _ConnectionLostException: If we can't seem to talk to the machine.
    _StopException: If a stop was requested.
    _ProtocolViolationException: If the responses packet violates the protocol.

    """
    request_action = _SHORT_STRUCT.unpack(packet[4:6])[0]
    for attempt in range(max_tries):
        _write_to_port(port, packet)
        try:
            response = _read_packet(port, stop, buf)
            if response[1] != packet[1]:
                continue  # Wrong sequence number.
            response_action = _SHORT_STRUCT.unpack(response[4:6])[0]
            if request_action != response_action:
                raise _ProtocolViolationException()
            return response
        except _TimeoutException:
            continue
    raise _ConnectionLostException()


class _SequenceCounter:
    """A mod 256 counter."""

    def __init__(self, seq=0):
        """Init a new counter starting at seq."""
        self.seq = seq

    def __call__(self):
        """Return the next value."""
        cur, self.seq = self.seq, (self.seq + 1) % 256
        return cur


def _read(port, stop, seq, request_buf, response_buf, stroke_buf, block, byte):
    """Read the full contents of the current file from beginning to end.

    The file should be opened first.

    Args:
    - port: The port to use.
    - stop: The event used to request stopping.
    - seq: A _SequenceCounter instance to use to track packets.
    - request_buf: Buffer to use for request packet.
    - response_buf: Buffer to use for response packet.
    - stroke_buf: Buffer to use for strokes read from the file.

    Raises:
    _ProtocolViolationException: If the protocol is violated.
    _StopException: If a stop is requested.
    _ConnectionLostException: If we can't seem to talk to the machine.

    """
    bytes_read = 0
    while True:
        packet = _make_read(request_buf, seq(), block, byte, length=512)
        response = _send_receive(port, stop, packet, response_buf)
        p1 = _SHORT_STRUCT.unpack(response[8:10])[0]
        if not (
            (p1 == 0 and len(response) == 14)  # No data.
            or (p1 == len(response) - 16)
        ):  # Data.
            raise _ProtocolViolationException()
        if p1 == 0:
            return block, byte, buffer(stroke_buf, 0, bytes_read)
        data = buffer(response, 14, p1)
        _write_to_buffer(stroke_buf, bytes_read, data)
        bytes_read += len(data)
        byte += p1
        if byte >= 512:
            block += 1
            byte -= 512


def _loop(port, stop, callback, ready_callback, timeout=1):
    """Enter into a loop talking to the machine and returning strokes.

    Args:
    - port: The port to use.
    - stop: The event used to signal that it's time to stop.
    - callback: A function that takes a list of pressed keys, called for each
    stroke.
    - ready_callback: A function that is called when the machine is ready.
    - timeout: Timeout to use when waiting for a response in seconds. Should be
    1 when talking to a real machine. (default: 1)

    Raises:
    _ProtocolViolationException: If the protocol is violated.
    _StopException: If a stop is requested.
    _ConnectionLostException: If we can't seem to talk to the machine.

    """
    # We want to give the machine a standard timeout to finish whatever it's
    # doing but we also want to stop if asked to so this is the safe way to
    # wait.
    if stop.wait(timeout):
        raise _StopException()
    port.flushInput()
    port.flushOutput()
    # Set serial port timeout to the timeout value
    port.timeout = timeout
    # With Python 3, our replacement for buffer(), using memoryview, does not
    # allow resizing the original bytearray(), so make sure our buffers are big
    # enough to begin with.
    request_buf, response_buf = _allocate_buffer(), _allocate_buffer()
    stroke_buf = _allocate_buffer()
    seq = _SequenceCounter()
    request = _make_open(request_buf, seq(), b"A", b"REALTIME.000")
    # Any checking needed on the response packet?
    _send_receive(port, stop, request, response_buf)
    # Do a full read to get to the current position in the realtime file.
    block, byte = 0, 0
    block, byte, _ = _read(
        port, stop, seq, request_buf, response_buf, stroke_buf, block, byte
    )
    ready_callback()
    while True:
        block, byte, data = _read(
            port, stop, seq, request_buf, response_buf, stroke_buf, block, byte
        )
        strokes = _parse_strokes(data)
        for stroke in strokes:
            callback(stroke)


class Stentura(plover.machine.base.SerialStenotypeBase):
    """Stentura interface.

    This class implements the three methods necessary for a standard
    stenotype interface: start_capture, stop_capture, and
    add_callback.
    """

    KEYS_LAYOUT = """
        #  #  #  #  #  #  #  #  #  #
        S- T- P- H- * -F -P -L -T -D
        S- K- W- R- * -R -B -G -S -Z
              A- O-   -E -U
        ^
    """

    def run(self):
        """Overrides base class run method. Do not call directly."""
        try:
            _loop(self.serial_port, self.finished, self._notify_keys, self._ready)
        except _StopException:
            pass
        except Exception:
            log.info("Failure starting Stentura", exc_info=True)
            self._error()
